---@class Component : Object
---@field protected __class string
---@field private _parent Component
---@field private _root LuaGuiElement
---@field private _children Component[]
Component = Object:extendAs("model.Component")

---[STATIC]
---Sets element.tags.className to value of self.__className to recognize it while searching
---@protected
---@generic C : Component
---@param class C
---@param parent Component
---@param addParameters LuaGuiElement.add_parameters
---@return C
function Component.create(class, parent, addParameters)
    addParameters.tags = { className = class:className() }
    local instance = class:new(parent, parent:root().add(addParameters))
    parent:addChild(instance)
    return instance
end

---[STATIC]
---@protected
---@param parent Component
---@param root LuaGuiElement
---@return self
function Component:new(parent, root)
    local this = Component:super(Object:new())
    this._parent = assert(parent)
    this._root = assert(root)
    this._children = {}
    return this
end

---@private
---@param child Component
function Component:addChild(child)
    table.insert(self._children, child)
end

function Component:removeChild(childToRemove)
    for i, child in ipairs(self._children) do
        if child:root() == childToRemove:root() then
            table.remove(self._children, i)
        end
    end
end

---@public
---@param click GuiClickEvent
function Component:handleGuiClick(click)
    local child = self:findChildContaining(click:element())
    if child then
        child:handleGuiClick(click)
    end
end

---@public
---@param guiChangedEvent GuiEvent
function Component:handleGuiChange(guiChangedEvent)
    local child = self:findChildContaining(guiChangedEvent:element())
    if child then
        child:handleGuiChange(guiChangedEvent)
    end
end

---@public
---@param event Event
function Component:handle(event)
    for _, child in ipairs(self:children()) do
        child:handle(event)
    end
end

---@public
function Component:lock()
    for _, child in ipairs(self:children()) do
        child:lock()
    end
end

---@public
function Component:unlock()
    for _, child in ipairs(self:children()) do
        child:unlock()
    end
end

---@return self[]
function Component:siblingsAndMe()
    return self:parent():childrenOfType(self:class())
end

---@generic T : Component
---@param class T
---@return T
function Component:childOfType(class)
    return self:childrenOfType(class)[1]
end

---@generic T : Component
---@param class T
---@return T[]
function Component:childrenOfType(class)
    local children = {}
    for _, child in ipairs(self:children()) do
        if child:isInstanceOf(class) then
            table.insert(children, child)
        end
    end
    return children
end

---@public
---@return Component[]
function Component:children()
    local children = {}
    for _, child in ipairs(self._children) do
        table.insert(children, child)
    end
    return children
end

---@protected
---@return Component[]
function Component:loadChildren()
    self._children = {}
    for _, class in ipairs(self:childrenClasses()) do
        for _, child in ipairs(self:findChildOfClass(class)) do
            table.insert(self._children, child)
        end
    end
end

---@return Component[]
function Component:childrenClasses()
    return {}
end

---@private
---@generic T : Component
---@param childClass T
---@return T[]
function Component:findChildOfClass(childClass)
    if self:root().children then
        local found = {}
        for _, childRoot in ipairs(self:root().children) do
            if childClass:isRepresentedBy(childRoot) then
                table.insert(found, childClass:new(self, childRoot))
            end
        end
        return found
    else
        return {}
    end
end

---[STATIC]
---@public
---@param element LuaGuiElement
---@return boolean
function Component:isRepresentedBy(element)
    return element.tags and element.tags.className == self:className()
end

---@generic T : Component
---@param class T
---@return T
function Component:ancestor(class)
    if class:isInstanceOf(Screen) then
        error("No ancestor of class: " .. class:className())
    elseif self:isInstanceOf(class) then
        return self
    else
        return self:parent():ancestor(class)
    end
end

---@protected
---@param replacement Component
function Component:replaceWith(replacement)
    self:parent():root().swap_children(self:index(), replacement:index())
    self:delete()
end

---@public
---@return boolean
function Component:isValid()
    return self:root().valid
end

---@public
function Component:show()
    self:root().visible = true
end

---@public
function Component:hide()
    self:root().visible = false
end

---@public
---@return boolean
function Component:isVisible()
    return self:root().visible
end

---@public
function Component:delete()
    self:parent():removeChild(self)
    self:root().destroy()
end

---@protected
---@return number
function Component:index()
    return self:root().get_index_in_parent()
end

---@protected
---@return Component
function Component:parent()
    return self._parent
end

---@protected
---@return LuaGuiElement
function Component:parentElement()
    return self:root().parent
end

---@protected
---@return LuaGuiElement
function Component:root()
    return self._root
end

---@public
---@return boolean
---@param guiEvent GuiEvent
function Component:isOriginOf(guiEvent)
    return self:root() == guiEvent:element()
end

---@private
---@param element LuaGuiElement
---@return Component
function Component:findChildContaining(element)
    for _, child in ipairs(self:children()) do
        if child:contains(element) then
            return child
        end
    end
end

---@private
---@param element LuaGuiElement
---@return boolean
function Component:contains(element)
    if element.parent == nil then
        return false
    elseif self:root() == element then
        return true
    else
        return self:contains(element.parent)
    end
end
