import('model.events.gui.GuiLocationChangedEvent')
import('model.Component')
import('model.toolbar.header.ToolbarHeader')
import('model.toolbar.header.Unlock')
import('model.toolbar.content.ToolbarContent')
import('model.toolbar.content.sections.Sections')
import('model.toolbar.content.sections.section.Section')
import('model.toolbar.content.sections.section.content.table.Table')

---@class Toolbar : Component
Toolbar = Component:extendAs("model.toolbar.Toolbar")

function Toolbar:create(parent)
    local instance = Component.create(self, parent, {
        type = "frame",
        direction = "vertical",
        style = "toolbar" })
    ToolbarHeader:create(instance)
    ToolbarContent:create(instance)
    instance:adjustSize()
    return instance
end

function Toolbar:new(parent, root)
    local instance = Toolbar:super(Component:new(parent, root))
    instance:loadChildren()
    return instance
end

function Toolbar:childrenClasses()
    return { ToolbarHeader, ToolbarContent }
end

---@return ToolbarHeader
function Toolbar:header()
    return self:childOfType(ToolbarHeader)
end

---@return ToolbarContent
function Toolbar:content()
    return self:childOfType(ToolbarContent)
end

---@private
---@return boolean
function Toolbar:isLocked()
    return self:childOfType(ToolbarHeader):childOfType(Unlock) ~= nil
end

function Toolbar:handleGuiChange(guiChangedEvent)
    if self:isOriginOf(guiChangedEvent) then
        if guiChangedEvent:isInstanceOf(GuiLocationChangedEvent) then
            self:keepWithinTheScreen()
            self:turnOffCenteringOnScreen()
        end
    else
        Component.handleGuiChange(self, guiChangedEvent)
    end
end

---@private
function Toolbar:keepWithinTheScreen()
    local location = self:location()
    if location.x < 0 then
        location.x = 0
    end
    if location.y < 0 then
        location.y = 0
    end
    self:setLocation(location)
end

---@private
---@return GuiLocation
function Toolbar:location()
    return self:root().location
end

---@private
---@param location GuiLocation
function Toolbar:setLocation(location)
    self:root().location = location
end

---@private
function Toolbar:turnOffCenteringOnScreen()
    self:root().auto_center = false
end

---@public
function Toolbar:centerOnScreen()
    self:root().force_auto_center()
end

---@public
function Toolbar:collapse()
    self:content():hide()
end

---@public
function Toolbar:expand()
    self:content():show()
end

---@public
function Toolbar:addSection()
    self:content():sections():addSection()
    self:adjustSize()
end

---@public
---@return number
function Toolbar:sectionsCount()
    return self:content():sections():count()
end

function Toolbar:lock()
    self:content():root().style = "toolbar_content"
    Component.lock(self)
    self:adjustSize()
end

function Toolbar:unlock()
    Component.unlock(self)
    self:adjustSize()
end

---@public
function Toolbar:adjustSize()
    if self:isLocked() then
        self:trim()
    else
        self:resize(4, 1, 1)
    end
end

---@private
function Toolbar:trim()
    self:resize(1, 1, 0)
end

---@private
function Toolbar:resize(minimumWidth, minimumHeight, bottomPadding)
    local width = math.max(minimumWidth, self:lastOccupiedColumnIndex() + bottomPadding)
    self:resizeTables(width, minimumHeight, bottomPadding)
end

---@private
---@return number
function Toolbar:lastOccupiedColumnIndex()
    local lastOccupiedColumnIndex = 0
    for _, table in ipairs(self:tables()) do
        lastOccupiedColumnIndex = math.max(lastOccupiedColumnIndex, table:lastOccupiedColumnIndex())
    end
    return lastOccupiedColumnIndex
end

---@private
---@param width number
function Toolbar:resizeTables(width, minimumHeight, bottomPadding)
    for _, table in ipairs(self:tables()) do
        table:resize(width, minimumHeight, bottomPadding)
    end
end

---@private
---@return Table[]
function Toolbar:tables()
    local tables = {}
    for _, section in pairs(self:content():sections():list()) do
        if section:isExpanded() then
            table.insert(tables, section:content():table())
        end
    end
    return tables
end
