import('model.events.CursorStackChanged')
import('model.events.InventoryChanged')
import('model.toolbar.content.sections.section.content.table.Slot')

---@class Item : Slot
Item = Slot:extendAs("model.toolbar.content.sections.section.content.table.slots.Item")

function Item:create(parent, itemName)
    local instance = Component.create(self, parent, {
        type = "sprite-button",
        elem_tooltip = { type = "item", name = itemName },
        sprite = "item/" .. assert(itemName),
        style = "toolbar_content_sections_section_content_table_row_slot_item",
        mouse_button_filter = { "left", "middle", "right", "button-5" },
    })
    instance:updateCount()
    return instance
end

function Item:new(parent, root)
    return Item:super(Slot:new(parent, root))
end

function Item:handleGuiClick(click)
    if click:isLeft() and not click:withModifier() then
        self:handlePick()
    elseif click:isMiddle() and not click:withModifier() then
        self:handleClear()
    elseif (click:isLeft() or click:isRight() or click:isButton5()) and click:withModifier() then
        self:handleCraft(click)
    end
end

function Item:handlePick()
    local itemName = self:itemName()
    if self:inventory():hasItem(itemName) then
        self:inventory():pick(itemName)
    else
        self:cursor():pickGhost(itemName)
    end
    self:rememberPick()
    self:notifyAboutSuccess()
end

function Item:handleClear()
    self:clear()
    self:notifyAboutSuccess()
end

---@param click GuiClickEvent
function Item:handleCraft(click)
    if not self:player().character then
        self:notifyAboutFailure("gui-map-editor.title")
    elseif self:hasRecipe() then
        local recipe = self:recipe()
        if not recipe.enabled then
            self:notifyAboutFailure("description.recipe-not-researched")
        elseif self:player().force.get_hand_crafting_disabled_for_recipe(recipe) then
            self:notifyAboutFailure("recipe-not-craftable-in-hand")
        elseif self:player().get_craftable_count(recipe) == 0 then
            self:notifyAboutFailure("not-enough-ingredients")
        else
            local requestedCount = self:recognizeRequestedCount(click)
            if requestedCount then
                local startedCount = self:player().begin_crafting { count = requestedCount, recipe = recipe.name }
                self:notifyAboutSuccess(startedCount)
            else
                self:notifyAboutFailure()
            end
        end
    else
        self:notifyAboutFailure("entity-status.no-recipe")
    end
end

---@private
---@return boolean
function Item:hasRecipe()
    return self:recipe() ~= nil
end

---@private
---@return LuaRecipe
function Item:recipe()
    local recipePrototype = self:recipePrototype()
    if recipePrototype then
        return self:player().force.recipes[recipePrototype.name]
    end
end

---@private
---@return LuaRecipePrototype
function Item:recipePrototype()
    local recipes = game.get_filtered_recipe_prototypes(
            { {
                  filter = "has-product-item",
                  elem_filters = { { filter = "name", name = self:itemName() } }
              } })
    for _, recipe in pairs(recipes) do
        return recipe
    end
    return nil
end

---@param click GuiClickEvent
function Item:recognizeRequestedCount(click)
    if click:isLeft() and click:withOnlyCtrl() then
        return 1
    elseif (click:isRight() or click:isButton5()) and click:withOnlyCtrl() then
        return 5
    elseif click:isLeft() and click:withOnlyAlt() then
        return self:itemPrototype().stack_size
    elseif (click:isRight() or click:isButton5()) and click:withOnlyAlt() then
        return self:itemPrototype().stack_size / 2
    elseif (click:isRight() or click:isButton5()) and (click:withOnlyShift() or (click:withCtrl() and click:withAlt() and not click:withShift())) then
        return self:player().get_craftable_count(self:recipe()) / 2
    elseif click:isLeft() and (click:withOnlyShift() or (click:withCtrl() and click:withAlt() and not click:withShift())) then
        return self:player().get_craftable_count(self:recipe())
    else
        return nil
    end
end

---@private
---@return LuaItemPrototype
function Item:itemPrototype()
    return game.item_prototypes[self:itemName()]
end

---@private
---@param message string | nil
function Item:notifyAboutSuccess(message)
    self:player().play_sound { path = "utility/inventory_click" }
    if message then
        self:player().create_local_flying_text { text = message, create_at_cursor = true }
    end
end

---@private
---@param message string | nil
function Item:notifyAboutFailure(message)
    self:player().play_sound { path = "utility/cannot_build" }
    if message then
        self:player().create_local_flying_text { text = { message }, create_at_cursor = true }
    end
end

---@param event Event
function Item:handle(event)
    if event:isInstanceOf(CursorStackChanged) then
        self:highlightHeldItem()
        self:updateCount()
    elseif event:isInstanceOf(InventoryChanged) then
        self:updateCount()
    end
end

---@private
function Item:highlightHeldItem()
    local cursor = self:cursor()
    if cursor:holdsStack() and cursor:itemName() == self:itemName() then
        self:root().style = "toolbar_content_sections_section_content_table_row_slot_item_selected"
    elseif cursor:holdsGhost() and cursor:itemName() == self:itemName() then
        self:root().style = "toolbar_content_sections_section_content_table_row_slot_item_selected"
    else
        self:root().style = "toolbar_content_sections_section_content_table_row_slot_item"
    end
end

---@private
function Item:updateCount()
    local count = self:inventory():countOf(self:itemName())
    local cursor = self:cursor()
    if cursor:holdsStack() and cursor:itemName() == self:itemName() then
        count = count + cursor:stackCount()
    end
    self:root().number = count > 0 and count or nil
end

function Item:itemName()
    return self:root().elem_tooltip.name
end
